#include "Fusion_SLAM_types.h"

#include "Sensor_CameraProcess.h"
#include "TueObjFusn_ObjectListType.h"
#include "TueObjFusn_ErrorCodes.h"
#include "tue_prv_objectSelection_int.h"
#include "MVSCamConverter_int.h"
#include "Fusion_SLAM_utils.h"
//#include "fusion_ext.h"

#ifndef TUE_CAMERA_RAW_OBJECT_NUM
#define TUE_CAMERA_RAW_OBJECT_NUM 24  // CAMERA object number get from sensor
#endif

/**
 * @defgroup group_camera_zones Camera preselection zones
 * @{
*/
#define  FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN         (0u)
#define  FUSION_PRV_OBJECTSELECTION_RADARONLY_BIN              (1u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE1_BIN               (2u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE2_BIN               (3u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE3_BIN               (4u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE4_BIN               (5u)

#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE1_BIN         (6u)
#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE2_BIN         (7u)
#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE3_BIN         (8u)
#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE4_BIN         (9u)

#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE1_BIN    (10u)
#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE2_BIN    (11u)
#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE3_BIN    (12u)
#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE4_BIN    (13u)


#define  FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI1_THRESHOLD_OBJ    (70.0f)
#define  FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI0_THRESHOLD_OBJ    (200.0f)
#define  FUSION_PRV_OBJECTSELECTION_ANGLE_ROI2_THRESHOLD_OBJ            (10.0f)
#define  FUSION_PRV_OBJECTSELECTION_ANGLE_ROI1_THRESHOLD_OBJ            (22.5f)
#define  FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE6               (50.0f)
#define  FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE5               (100.0f)
#define  FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE4               (150.0f)

/** @} */

#define FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD	4

/**
 * @def SENSOR_CAMERA2FUSION
 * 
 * @brief Defines the number of Camera objects that are preserved
 * for the fusion process.
*/
#define SENSOR_CAMERA2FUSION		5

/**
 * @def CAMERA_OBJECT_NUM
 * 
 * @brief Initial number of Camera objects from sensor
 * befor preselection
*/
#define CAMERA_OBJECT_NUM			(TUE_CAMERA_RAW_OBJECT_NUM)


extern uint32  Fusion_calculateObjectRange(float32 f32PosX, float32 f32PosY, float32* f32Range);
extern uint32	Fusion_calculateObjectAngle(float32 f32PosX, float32 f32PosY, float32* f32Angle);


/** 
 * @brief Bin position of each camera track is set here
 *
 * Based on track distance from origin and the angle, it's assigned to one of
 * the 14 available zones. This function fills the corresponding bin of
 * the input track. For available zones and their index, refer to @ref group_camera_zones
 * 
 * @param[in] pSensorObject Input camera track
 * @param[out] as16TrkblBinCounter Array holding number of tracks in each zone (bin)
 * @param[out] fRange Track range which is calculated during the function call
 * 
 * @return Track bin index
*/
uint8 Fusion_setCameraBinPosition(CamObject* pSensorObject, sint16 as16TrkblBinCounter[], float32 *fRange)
{
	float32 f32ObjectRange;
	float32 f32ObjectAngle;
	uint32 u32Success;
	uint8 u8TrkblBinPosition = 0;
	VehDyn_t *pVehDyn	 = FusionInput_RawData;

	u32Success = Fusion_calculateObjectRange(pSensorObject->Kinematic.fDistX, pSensorObject->Kinematic.fDistY, &f32ObjectRange);
	u32Success |= Fusion_calculateObjectAngle(pSensorObject->Kinematic.fDistX, pSensorObject->Kinematic.fDistY, &f32ObjectAngle);
	BML_t_TrajRefPoint DistVEYPoint2Circle = CML_CalculateDistancePoint2Circle(
		pSensorObject->Kinematic.fDistX,
		pSensorObject->Kinematic.fDistY,
		pVehDyn->Lateral.Curve.Curve);
	
                                         
	float32 f_Dist2TrajAbs = BML_f_Abs(DistVEYPoint2Circle.f_DistToTraj);

	*fRange = f32ObjectRange;

	 uint8  ObjectType = pSensorObject->Classification.ObjectType;
	
	   if (((FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI1_THRESHOLD_OBJ < f32ObjectRange) &&	(FUSION_PRV_OBJECTSELECTION_ANGLE_ROI2_THRESHOLD_OBJ < f32ObjectAngle)) || 
		   ((FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI1_THRESHOLD_OBJ > f32ObjectRange) &&	(FUSION_PRV_OBJECTSELECTION_ANGLE_ROI1_THRESHOLD_OBJ < f32ObjectAngle)) ||  
			(FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI0_THRESHOLD_OBJ < f32ObjectRange)) 
	   {
			u8TrkblBinPosition = FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN;
			as16TrkblBinCounter[FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN]++;
	   } /* Remove all radar only stationary objects */
	   else if (FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE4 < f32ObjectRange)
	   {
		   if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
		   {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE1_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE1_BIN]++;
		   	} else {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE1_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE1_BIN]++;
			} 
	   }
	   else if ((FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE4 > f32ObjectRange) && 
				(FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE5 < f32ObjectRange))
	   {	   

			 if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
			 {
				  u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE2_BIN;
				  as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE2_BIN]++;
			  } else {
				  u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE2_BIN;
				  as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE2_BIN]++;
			  }
	   }
	   else if ((FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE5 > f32ObjectRange) && 
				(FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE6 < f32ObjectRange))
	   {	   

			if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
			{
				 u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE3_BIN;
				 as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE3_BIN]++; 		 
			
			 } else {
				 u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE3_BIN;
				 as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE3_BIN]++;
			 }
	   }
	   else if (FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE6 > f32ObjectRange)
	   { 
		 if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
		 {
			   u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE4_BIN;
			   as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE4_BIN]++;
		  } else {
			  u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE4_BIN;
			  as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE4_BIN]++;
		  }

	   }
	   else
	   {
		  /* MISRA */
		   u8TrkblBinPosition = FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN;
		   as16TrkblBinCounter[FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN]++;

	   }
	
	return u8TrkblBinPosition;
}


/**
 * @brief Sets which camera tracks must be dropped
 * 
 * This function specifies the last bin index to which tracks must be
 * deleted. It iterates through track bin indices and eliminates as many
 * tracks as needed so that the number of tracks to be desired.
 * 
 * @param[in] s16NumberTrackablesToDelete Number of tracks that must be deleted
 * @param[in] as16TrkblBinCounter Array specifying number of tracks in each bin
 * @param[out] u8LastBinToDelete Index of the last bin to which tracks must be deleted
 * 
 * @return error status
*/
static uint32  Fusion_setDroppedCameraBins(uint8* u8LastBinToDelete, sint16 s16NumberTrackablesToDelete, sint16 as16TrkblBinCounter[])
{
   uint32 u32Success = TUEOBJFUSN_ERROR_NOERROR;   /* PRQA S 3204 1 *//* variable may be set to other values depending on run-time check activation */
   uint8 u8Bin;
   uint8 u8LastBin;

   u8LastBin = TUE_PRV_OBJECTSELECTION_MAX_BINS;
   for (u8Bin = 0u; (u8Bin < TUE_PRV_OBJECTSELECTION_MAX_BINS) && (0 < s16NumberTrackablesToDelete); u8Bin++)
   {
      if (as16TrkblBinCounter[u8Bin] < s16NumberTrackablesToDelete)
      {  
         /* bin has to be dropped completly -> sets bin to be dropped (TUE_PRV_OBJECTSELECTION_DROP_BIN)*/
         s16NumberTrackablesToDelete -= as16TrkblBinCounter[u8Bin];
      }
      else
      {  
         /* bin has to be dropped party -> write number which has to be dropped */ 
         s16NumberTrackablesToDelete = 0;
         u8LastBin = u8Bin; /* this is the last bin which has to be (partly) deleted */
      }
   }
   *u8LastBinToDelete = u8LastBin;


   return u32Success;
}


/**
 * @brief Copies preserved camera tracks and convert the type
 * 
 * A boolean list determining which elements must be dropped is 
 * provided as input and this function only keeps a subset of tracks
 * based on the array. Also the type of each track is converted to fusion 
 * core type.
 * 
 * @param[in] pSensorObjList original camera object list
 * @param[in] abObjectsToBeDropped array showing which tracks must be deleted
 * @param[out] pTueObjList converted camera object list
 * 
 * @return Error status
 * 
*/
uint32 Fusion_CameracopyLkfTrkbl2Outputlist(CamObjectList* pSensorObjList, 
											const boolean abObjectsToBeDropped[], 
											TueObjFusn_ObjectListType* pTueObjList)
{
   uint32 u32Success = TUEOBJFUSN_ERROR_NOERROR;
   uint16 u16i;
   uint16 u16NumObj      = 0u;
   sint16 s16Idx;

   CamObject*	pSenObject;
   TueObjFusn_TrackableType*		pInterObject;

   for (u16i = 0u; u16i < CAMERA_OBJECT_NUM && u16NumObj < SENSOR_CAMERA2FUSION; u16i++)
   {
   
		pSenObject = &(pSensorObjList->aObject[u16i]);
      if ((FALSE == abObjectsToBeDropped[u16i]) && (pSenObject->Kinematic.fDistX > 0.1))
      {
         pInterObject = &(pTueObjList->aTrackable[u16NumObj]);

		 //LogWrite("MVSObjToTrackableConverter_convertObject %d, %d, %.02f, %.02f", u16i, u16NumObj, pSenObject->Kinematic.fDistX, pSenObject->Kinematic.fDistY);

		 MVSObjToTrackableConverter_convertObject(pInterObject, pSenObject);

         u16NumObj += 1u;
      }
      else
      {
         /* MISRA */
      }
   }  /* end for u16i */
   pTueObjList->u16NumObjects = u16NumObj;

   return u32Success;
}



/** 
 * @brief Camera object preselection
 *
 * Specifies which camera objects to be preserved and which to be 
 * thrown away.
 * 
 * @param[in] pSenObjList list of all camera objects
 * @param[out] abObjectsToBeDropped boolean list determining which objects to throw away
 *
 * @return none
 * 
*************************************************************/
void Fusion_prv_Cameraobject_select(CamObjectList* pSenObjList, boolean abObjectsToBeDropped[])
{
	uint32 u32Success;
	uint8 u8LastBinToDelete;
	uint16 u16i, u8Bin;
	CamObject *pSenObject;
	uint8 auBinPosition[CAMERA_OBJECT_NUM]= {0};
	sint16  as16CameraBinCounter[14] = {0};
	uint8 iNumOfUsedObjects = 0;
	float32 Dist2TrajAbsArray[BSW_OBJECT_NUM_MAX] = {0};
	float32 f32ObjectRange;
	uint8 i, j;
	sint16 s16ObjectsToDelete = 0;

	for(u16i = 0; u16i < CAMERA_OBJECT_NUM; u16i++) {
		if(pSenObjList->aObject[u16i].Kinematic.fDistX > 0.01) {
			iNumOfUsedObjects++;
		}
	}

	if(iNumOfUsedObjects > SENSOR_CAMERA2FUSION) {
		s16ObjectsToDelete = iNumOfUsedObjects - SENSOR_CAMERA2FUSION;
	}

	//LogWrite("Fusion_prv_Cameraobject_select: %d, %d", iNumOfUsedObjects, s16ObjectsToDelete);


	for(u16i = 0u; u16i < CAMERA_OBJECT_NUM; u16i++)
	{
	   abObjectsToBeDropped[u16i] = FALSE;
	}

	// pExtObjList->HeaderObjList.iNumOfUsedObjects
	for (u16i = 0u; u16i < CAMERA_OBJECT_NUM; u16i++)
	{
	   pSenObject = &pSenObjList->aObject[u16i];
	   auBinPosition[u16i] = 0xFF;
	   if(pSenObject->ObjectID != 0xFF) {
			auBinPosition[u16i] = Fusion_setCameraBinPosition(pSenObject, as16CameraBinCounter, &f32ObjectRange);
			Dist2TrajAbsArray[u16i] = f32ObjectRange;
	   }
	}

	if(as16CameraBinCounter[0] > s16ObjectsToDelete) {
		s16ObjectsToDelete = as16CameraBinCounter[0];
	}

	Fusion_setDroppedCameraBins(&u8LastBinToDelete, s16ObjectsToDelete, as16CameraBinCounter);
	/* loop through bins and delete trackables */
	for (u8Bin = 0u; u8Bin <= u8LastBinToDelete; u8Bin++) 
	{ 
	   /* loop through all trackables and delete which are in this bin */
		if(s16ObjectsToDelete >= as16CameraBinCounter[u8Bin]) {
			   for (u16i = 0u; (u16i < TUE_CAMERA_RAW_OBJECT_NUM) && (0 < s16ObjectsToDelete); u16i++)
			   {
		//		  pExtObject = &pExtObjList->Objects[u16i];;
				  /* Remove coasted stationary objects */
				  if (u8Bin == auBinPosition[u16i])  
				  {
					 s16ObjectsToDelete--;
					 abObjectsToBeDropped[u16i] = TRUE;
					 
//					 LogWrite("Fusion_prv_Cameraobject_select 3 abObjectsToBeDropped :%d, %.02f, %.02f, %.02f",
//					 	u16i,
//					 	Dist2TrajAbsArray[u16i],
//					 	pSenObjList->aObject[u16i].Kinematic.fDistX,
//					 	pSenObjList->aObject[u16i].Kinematic.fDistY);
				  }   
				  else
				  {
					 /* MISRA */
				  }
			   }  
		} else {
			break;
		}
	}

	if(s16ObjectsToDelete > 0) {
			sint8 auBinIdx[CAMERA_OBJECT_NUM]= {0};
			uint8 sobjNum = 0;
			sint8 arrayIndex = -1;
			for(i = 0; i < CAMERA_OBJECT_NUM; i++) {
				if(auBinPosition[i] == u8Bin) {
					auBinIdx[sobjNum] = i;
					sobjNum++;
				}
			}
		
			for(i = 0; i < s16ObjectsToDelete; i++) {
				float32 tempMinDist = -1;
				for(j = 0; j < sobjNum; j++) {
					if (Dist2TrajAbsArray[auBinIdx[j]] >= 0)
					{
						if (BML_f_Abs(tempMinDist + 1) < FLT_EPSILON ||   Dist2TrajAbsArray[auBinIdx[j]] > tempMinDist)
						{
							tempMinDist = Dist2TrajAbsArray[auBinIdx[j]];
							arrayIndex = auBinIdx[j];
						}
					}
	
				}
				if (arrayIndex >= 0 && arrayIndex < CAMERA_OBJECT_NUM )
				{
					Dist2TrajAbsArray[arrayIndex] = -1;
					abObjectsToBeDropped[arrayIndex] = TRUE;
//					LogWrite("Fusion_prv_Cameraobject_select 4 abObjectsToBeDropped :%d, %.02f", arrayIndex, Dist2TrajAbsArray[u16i]);
//					LogWrite("Fusion_prv_Cameraobject_select 4 abObjectsToBeDropped :%d, %.02f, %.02f, %.02f",
//					   arrayIndex,
//					   Dist2TrajAbsArray[u16i],
//					   pSenObjList->aObject[u16i].Kinematic.fDistX,
//					   pSenObjList->aObject[u16i].Kinematic.fDistY);


				} else {
					LogWrite("Fusion_prv_Cameraobject_select ERROR ");
	
				}
			}
	
		}

}

/** 
 * @brief Selects a subset of Camera objects to be processed and converts types
 *
 * Only a subset of Camera objects are selected to be processed and fused. 
 * Specifically, #SENSOR_CAMERA2FUSION defines how many to keep.
 * 
 * @param[in] pSenObjList list of all Camera objects
 * @param[in] u32globalTimeStamp global system timestamp
 * @param[out] pTueObjList list of preselected camera objects
 *
 * @return none
 * 
*/
void Fusion_CameraObjectPreSelection(CamObjectList* pSenObjList, 
									 TueObjFusn_ObjectListType* pTueObjList, 
									 uint32 u32globalTimeStamp)
{
	boolean abObjectsToBeDropped[CAMERA_OBJECT_NUM] = {0};
	boolean  bObjectIsValid = TRUE;
	static uint16 u16UpdateCount = 0u;
	static uint32 u32CamTimeStampPrev = 0u;
	float32  f32CorrLatency = FLT_ZERO;
	uint16   u16NumConvertedObjects = 0u;


	if (u32CamTimeStampPrev != pSenObjList->sSigHeader.uiTimeStamp)
	{
		u16UpdateCount = ((u16UpdateCount + 1u) % TUEOBJFUSN_OBJECTLIST_U16LISTUPDATECOUNTER_MAX);
		u32CamTimeStampPrev = (pSenObjList->sSigHeader.uiTimeStamp);
	}
	else
	{
		/* MISRA */
	}

	pTueObjList->u16NumObjects = 0u;
	pTueObjList->u16ListUpdateCounter = u16UpdateCount;
	
	/* correct latency using global timestamp */
	if (u32globalTimeStamp >= u32CamTimeStampPrev)
	{
		f32CorrLatency = (float32)(u32globalTimeStamp - u32CamTimeStampPrev);
	}
	else
	{
		f32CorrLatency = (float32)(0xFFFFFFFF - u32CamTimeStampPrev + u32globalTimeStamp);
	}
	f32CorrLatency = MAX(0, MIN(1000000, (f32CorrLatency)));
	// 0.02: Default sensor latency
	pTueObjList->f32MeasurementLatency = (f32CorrLatency) / 1000.0 / 1000.0 + 0.02; //convert latency from us to s
	
	pTueObjList->u32SensorPattern = TUEOBJFUSN_TRACKABLE_U32SENSOR_MVS_FRONT;
	

	Fusion_prv_Cameraobject_select(pSenObjList, abObjectsToBeDropped);
	Fusion_CameracopyLkfTrkbl2Outputlist(pSenObjList, abObjectsToBeDropped, pTueObjList);
}

